﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Diagnostics;
using Microsoft.Xna.Framework;

namespace BasicSetupDemo
{
    public class QuaternionCamera
    {
        /// <summary>
        /// Gets and sets the current location of the camera.
        /// </summary>
        public Vector3 Position { get; set; }
        /// <summary>
        /// Gets and sets the current orientation of the camera.
        /// </summary>
        public Quaternion Orientation { get; set; }
        /// <summary>
        /// Gets the current world transform of the camera, including both orientation and translation.
        /// </summary>
        public Matrix WorldMatrix { get; private set; }
        /// <summary>
        /// Gets the current view matrix of the camera.
        /// </summary>
        public Matrix ViewMatrix { get; private set; }
        /// <summary>
        /// Projection matrix of the camera.
        /// </summary>
        public Matrix ProjectionMatrix { get; set; }
        /// <summary>
        /// Accumulated orientation change since the last update.
        /// </summary>
        private Vector3 orientationChange = Vector3.Zero;

        /// <summary>
        /// Constructs a new quaternion camera.
        /// </summary>
        /// <param name="position">Initial position of the camera.</param>
        /// <param name="orientation">Initial orientation of the camera.</param>
        public QuaternionCamera(Vector3 position, Quaternion orientation)
        {
            this.Position = position;
            this.Orientation = orientation;
            ProjectionMatrix = Matrix.CreatePerspectiveFieldOfView(MathHelper.PiOver4, 4f / 3f, .1f, 10000.0f);
            Update(); //Call the update to complete some initialization.
        }

        /// <summary>
        /// Rotates the orienation incrementally based on a vector representing a change in orientation.
        /// </summary>
        /// <param name="orientationChange">Rotation axis with magnitude equal to the amount of rotation change</param>
        /// <returns>New orientation.</returns>
        public Quaternion RotateQuaternionIncrementally(Quaternion toRotate, Vector3 orientationChange)
        {
            Quaternion quaternionChange = (new Quaternion(orientationChange.X, orientationChange.Y, orientationChange.Z, 0) * .5f) * toRotate;
            toRotate.X += quaternionChange.X;
            toRotate.Y += quaternionChange.Y;
            toRotate.Z += quaternionChange.Z;
            toRotate.W += quaternionChange.W;
            toRotate.Normalize();
            return toRotate;
        }

        /// <summary>
        /// Sets the orientation, world matrix, and view matrix.
        /// </summary>
        public void Update()
        {
            Orientation = RotateQuaternionIncrementally(Orientation, orientationChange);
            orientationChange = Vector3.Zero; //'Used up' the angular velocity.
            Matrix orientationMatrix = Matrix.CreateFromQuaternion(Orientation);
            WorldMatrix = orientationMatrix * Matrix.CreateTranslation(Position);
            ViewMatrix = Matrix.CreateTranslation(-Position) * Matrix.Transpose(orientationMatrix);

        }

        /// <summary>
        /// Rotates the camera around its y axis.
        /// </summary>
        /// <param name="angle">Magnitude of rotation.</param>
        public void Yaw(float angle)
        {
            orientationChange += WorldMatrix.Up * angle;
        }

        /// <summary>
        /// Rotates the camera around its x axis.
        /// </summary>
        /// <param name="angle">Magnitude of rotation.</param>
        public void Pitch(float angle)
        {
            orientationChange += WorldMatrix.Right * angle;
        }

        /// <summary>
        /// Rotates the camera around its z axis.
        /// </summary>
        /// <param name="angle">Magnitude of rotation.</param>
        public void Roll(float angle)
        {
            orientationChange += WorldMatrix.Forward * angle;
        }


        /// <summary>
        /// Moves the camera along its local forward direction.
        /// </summary>
        /// <param name="distance">Distance to move the camera.</param>
        public void MoveForward(float distance)
        {
            Position += WorldMatrix.Forward * distance;
        }
        /// <summary>
        /// Moves the camera along its local right direction.
        /// </summary>
        /// <param name="distance">Distance to move the camera.</param>
        public void MoveRight(float distance)
        {
            Position += WorldMatrix.Right * distance;
        }
        /// <summary>
        /// Moves the camera along its local up direction.
        /// </summary>
        /// <param name="distance">Distance to move the camera.</param>
        public void MoveUp(float distance)
        {
            Position += WorldMatrix.Up * distance;
        }
    }
}
